<?php

namespace NoccyLabs\App;

use NoccyLabs\App\Attribute\AppInfo;
use ReflectionClass;
use ReflectionNamedType;

abstract class App {

    protected array $optionGroups;

    protected array $optionProps;

    private $customEntry = null;

    abstract protected function main(array $args);

    public static function boot(string|App $app, array $options=[], bool $debug=false): void
    {
        if (is_string($app)) {
            if (!class_exists($app)) {
                throw new \ErrorException("Invalid application class {$app}");
            }
            $app = new $app();
        }

        $args = self::parseOptions($app);

        if ($app->customEntry) {
            call_user_func($app->customEntry, ...$args);
            return;
        }
        $app->main($args);

    }

    private static function parseOptions(App $app)
    {
        $appReflection = new ReflectionClass($app);
        $optionProps = [];
        $optionGroups = [];

        foreach ($appReflection->getProperties() as $property) {
            $attrs = $property->getAttributes();
            $propinfo = (object)[
                'property' => $property,
                'name' => $property->getName(),
                'type' => $property->getType(),
                'default' => $property->hasDefaultValue() ? $property->getDefaultValue() : null,
                'option' => null,
                'group' => null,
            ];
            foreach ($attrs as $attr) {
                switch ($attr->getName()) {
                    case Attribute\AsOption::class:
                        $propinfo->option = $attr->newInstance();
                        $propinfo->group = $propinfo->option->group;
                        break;
                    case Attribute\OptionGroup::class;
                        $group = $attr->newInstance();
                        $groupName = $group->name;
                        $groupLabel = $group->label ?? ($optionGroups[$groupName]??$group->name);
                        $optionGroups[$groupName] = $groupLabel;
                        $propinfo->group = $groupName;
                        break;
                }
            }
            if ($propinfo->option) $optionProps[$property->getName()] = $propinfo;
        }
        
        $app->optionProps = $optionProps;
        $app->optionGroups = $optionGroups;

        $short = ""; $long = [];
        $shortMap = []; $longMap = [];
        foreach ($optionProps as $name=>$prop) {
            if ($prop->option->short) {
                $short .= $prop->option->short . ($prop->option->hasValue ? ":" : "");
                $shortMap[$prop->option->short] = $name;
            }
            if ($prop->option->long) {
                $long[] = $prop->option->long . ($prop->option->hasValue ? ":" : "");
                $longMap[$prop->option->long] = $name;
            }
            if ($envvar = $prop->option->envvar) {
                $env = getenv($envvar);
                if ($env) {
                    $prop->property->setAccessible(true);
                    if ($prop->option->hasValue) {
                        $prop->property->setValue($app, $env);
                    } else {
                        $prop->property->setValue($app, true);
                    }
                }
    
            }
        }

        $parsed = getopt($short, $long, $optind);
        
        foreach ($parsed as $opt=>$value) {
            if (strlen($opt)==1 && array_key_exists($opt,$shortMap)) {
                $prop = $optionProps[$shortMap[$opt]];
            } elseif (array_key_exists($opt, $longMap)) {
                $prop = $optionProps[$longMap[$opt]];
            } else {
                throw new \ErrorException("Unknown option: {$opt}");
            }

            if (!$prop->option->hasValue) {
                $value = true;
            }
            if ($prop->type instanceof ReflectionNamedType && $prop->type == "array") {
                if (!is_array($value)) {
                    $value = [ $value ];
                }
            }

            $prop->property->setAccessible(true);
            $prop->property->setValue($app, $value);

            if (is_callable($prop->option->entry)) {
                $app->customEntry = $prop->option->entry;
            } elseif (is_callable([$app,$prop->option->entry])) {
                $app->customEntry = [$app,$prop->option->entry];
            } elseif ($prop->option->entry) {
                throw new \LogicException("Invalid entrypoint for option {$prop->name}");
            }
        }

        return array_slice($GLOBALS['argv'], $optind);

    }
}
