<?php

namespace NoccyLabs\App\Trait\Process;

class Process
{

    private $proc;

    private bool $exited = false;

    private ?int $exitCode = null;

    private array $pipes = [];

    private array $messageQueue = [];

    public function __construct(
        public readonly string $processName,
        public readonly array $commandLine,

    )
    {
        $this->openProcess();
    }

    public function update(): void
    {
        // Make sure we have a valid process
        if (!is_resource($this->proc)) {
            $this->exited = true;
            return;
        }

        // Read process status
        $status = proc_get_status($this->proc);
        if (!$status['running'] && !$this->exited) {
            // Read any pending messages so we can close the stream
            $this->readMessagesFromProcess();
            $this->exited = true;
            $this->exitCode = $status['exitcode'];
        }

        /*
        pid       int    process id 
        running   bool   true if the process is still running, false if it has terminated. 
        signaled  bool   true if the child process has been terminated by an uncaught signal. Always set to false on Windows. 
        stopped   bool   true if the child process has been stopped by a signal. Always set to false on Windows. 
        exitcode  int    The exit code returned by the process (which is only meaningful if running is false). Only first call of this function return real value, next calls return -1. 
        termsig   int    The number of the signal that caused the child process to terminate its execution (only meaningful if signaled is true). 
        stopsig   int    The number of the signal that caused the child process to stop its execution (only meaningful if stopped is true).
        */
    }

    public function terminate(): void
    {

    }

    public function getStatus(): array
    {
        return [
            "name" => $this->processName,
            "command" => $this->commandLine,
            "running" => $this->isRunning(),
        ];
    }

    private function readMessagesFromProcess(): void
    {
        $p = $this->pipes[1]??null;
        if (!$p || !is_resource($p)) return;

        $r = [$p]; $w = []; $e = [];

        while (stream_select($r, $w, $e, 0)) {
            $read = fgets($r, 32000);
            if (str_starts_with($read,"{") && ($j = @json_decode($read))) {
                $this->messageQueue[] = $j;
            } else {
                $this->messageQueue[] = $read;
            }
        }
    }

    public function readMessage(): mixed
    {
        return array_shift($this->messageQueue);
    }

    public function sendMessage(mixed $message): void
    {
        $p = $this->pipes[0]??null;
        if (!$p || !is_resource($p)) return;

        fwrite($p, json_encode($message).PHP_EOL);
    }

    private function openProcess(): void
    {
        $ds = [
            0 => [ 'pipe', 'r' ],
            1 => [ 'pipe', 'w' ],
            2 => STDERR,
        ];
        $this->proc = proc_open($this->commandLine, $ds, $this->pipes);
    }

    public function isRunning(): bool
    {
        return !$this->exited;
    }

    public function isError(): bool
    {
        return ($this->exitCode != 0);
    }
}